/*
    Copyright (c) 2013 aizo ag, Zurich, Switzerland

    This file is part of libdsuid.

    libdsuid is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    libdsuid is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "dsuid.h"

#include <stdlib.h>
#include <stdio.h>
#include <uuid.h>

static int dsuid_from_uuid(const uuid_t *uuid, uint8_t index, dsuid_t *dsuid)
{
    uuid_rc_t rc;
    size_t data_len;
    unsigned char *str = NULL;

    rc = uuid_export(uuid, UUID_FMT_BIN, &str, &data_len);
    if ((rc != UUID_RC_OK) || (data_len != (DSUID_SIZE - 1))) {
        return DSUID_RC_UUID_ERR;
    }

    memcpy(dsuid->id, str, DSUID_SIZE);
    dsuid->id[16] = index;

    free(str);
    return DSUID_RC_OK;
}

static char *dsuid_get_ns_str(dsuid_ns_t ns)
{
    switch (ns) {
    case DSUID_NS_ENOCEAN:
        return "0ba94a7b-7c92-4dab-b8e3-5fe09e83d0f3";
    case DSUID_NS_GS1_128:
        return "8ca838d5-4c40-47cc-bafa-37ac89658962";
    case DSUID_NS_IEEE_MAC:
        return "ad019650-5330-4582-8506-62c7340b34e6";
    case DSUID_NS_VDSM:
        return "195de5c0-902f-4b71-a706-b43b80765e3d";
    case DSUID_NS_VDC:
        return "9888dd3d-b345-4109-b088-2673306d0c65";
    }
    return NULL;
}

int dsuid_generate_v3_from_namespace(dsuid_ns_t namespace,
                                     const char *name,
                                     dsuid_t *dsuid)
{
    uuid_t *uuid, *uuid_ns;
    uuid_rc_t rc;
    int rval;
    char *ns_str;

    if (name == NULL || strlen(name) == 0) {
        return DSUID_RC_WRONG_PAR;
    }

    ns_str = dsuid_get_ns_str(namespace);
    if (ns_str == NULL) {
        return DSUID_RC_WRONG_PAR;
    }

    rc = uuid_create(&uuid_ns);
    if (rc != UUID_RC_OK) {
        return DSUID_RC_UUID_ERR;
    }

    rc = uuid_create(&uuid);
    if (rc != UUID_RC_OK) {
        uuid_destroy(uuid_ns);
        return DSUID_RC_UUID_ERR;
    }

    rc = uuid_import(uuid_ns, UUID_FMT_STR, ns_str, UUID_LEN_STR);
    if (rc != UUID_RC_OK) {
        uuid_destroy(uuid);
        uuid_destroy(uuid_ns);
        return DSUID_RC_UUID_ERR;
    }

    rc = uuid_make(uuid, UUID_MAKE_V3, uuid_ns, name);
    if (rc != UUID_RC_OK) {
        uuid_destroy(uuid);
        uuid_destroy(uuid_ns);
        return DSUID_RC_UUID_ERR;
    }

    rval = dsuid_from_uuid(uuid, 0, dsuid);

    uuid_destroy(uuid_ns);
    uuid_destroy(uuid);
    return rval;
}

int dsuid_generate_v1(dsuid_t *dsuid)
{
    uuid_t *uuid;
    uuid_rc_t rc;
    int rval;

    rc = uuid_create(&uuid);
    if (rc != UUID_RC_OK) {
        return DSUID_RC_UUID_ERR;
    }

    rc = uuid_make(uuid, UUID_MAKE_V1);
    if (rc != UUID_RC_OK) {
        uuid_destroy(uuid);
        return DSUID_RC_UUID_ERR;
    }

    rval = dsuid_from_uuid(uuid, 0, dsuid);

    uuid_destroy(uuid);
    return rval;
}

int dsuid_generate_v1_mac(const char *mac, dsuid_t *dsuid)
{
    uuid_t *uuid;
    uuid_rc_t rc;
    int rval;

    rc = uuid_create(&uuid);
    if (rc != UUID_RC_OK) {
        return DSUID_RC_UUID_ERR;
    }

    rc = uuid_make(uuid, UUID_MAKE_V1);
    if (rc != UUID_RC_OK) {
        uuid_destroy(uuid);
        return DSUID_RC_UUID_ERR;
    }

    rval = dsuid_from_uuid(uuid, 0, dsuid);

    if (mac) {
        dsuid->id[10] = mac[0];
        dsuid->id[11] = mac[1];
        dsuid->id[12] = mac[2];
        dsuid->id[13] = mac[3];
        dsuid->id[14] = mac[4];
        dsuid->id[15] = mac[5];
    }

    uuid_destroy(uuid);
    return rval;
}

int dsuid_generate_v4_random(dsuid_t *dsuid)
{
    uuid_t *uuid;
    uuid_rc_t rc;
    int rval;

    rc = uuid_create(&uuid);
    if (rc != UUID_RC_OK) {
        return DSUID_RC_UUID_ERR;
    }

    rc = uuid_make(uuid, UUID_MAKE_V4);
    if (rc != UUID_RC_OK) {
        uuid_destroy(uuid);
        return DSUID_RC_UUID_ERR;
    }

    rval = dsuid_from_uuid(uuid, 0, dsuid);

    uuid_destroy(uuid);
    return rval;
}

const char *dsuid_strerror(int rc)
{
    switch ((dsuid_rc_t)rc) {
    case DSUID_RC_OK:
        return "dsuid: success";
    case DSUID_RC_INV_DSUID:
        return "dsuid: dsuid can not be used";
    case DSUID_RC_INV_SERIAL:
        return "dsuid: unexpected serial number";
    case DSUID_RC_GEN_FAIL:
        return "dsuid: general system failure";
    case DSUID_RC_WRONG_PAR:
        return "dsuid: wrong function paramater";
    case DSUID_RC_UUID_ERR:
        return "dsuid: error from ossp uuid lib";
    }
    return "dsuid: unknown return code";
}

const char *dsuid_rc_to_text(int rc)
{
    return dsuid_strerror(rc);
}

dsuid_t dsuid_from_dsid(const dsid_t *dsid)
{
    dsuid_t dsuid;

    dsuid.id[0] = dsid->id[0];
    dsuid.id[1] = dsid->id[1];
    dsuid.id[2] = dsid->id[2];
    dsuid.id[3] = dsid->id[3];
    dsuid.id[4] = dsid->id[4];
    dsuid.id[5] = dsid->id[5];
    dsuid.id[6] = 0;
    dsuid.id[7] = 0;
    dsuid.id[8] = 0;
    dsuid.id[9] = 0;
    dsuid.id[10] = dsid->id[6];
    dsuid.id[11] = dsid->id[7];
    dsuid.id[12] = dsid->id[8];
    dsuid.id[13] = dsid->id[9];
    dsuid.id[14] = dsid->id[10];
    dsuid.id[15] = dsid->id[11];
    dsuid.id[16] = 0;

    return dsuid;
}

int dsuid_to_dsid(const dsuid_t *dsuid, dsid_t *dsid)
{
    if ((dsuid->id[6] == 0) &&
        (dsuid->id[7] == 0) &&
        (dsuid->id[8] == 0) &&
        (dsuid->id[9] == 0))
    {
        dsid->id[0] = dsuid->id[0];
        dsid->id[1] = dsuid->id[1];
        dsid->id[2] = dsuid->id[2];
        dsid->id[3] = dsuid->id[3];
        dsid->id[4] = dsuid->id[4];
        dsid->id[5] = dsuid->id[5];
        dsid->id[6] = dsuid->id[10];
        dsid->id[7] = dsuid->id[11];
        dsid->id[8] = dsuid->id[12];
        dsid->id[9] = dsuid->id[13];
        dsid->id[10] = dsuid->id[14];
        dsid->id[11] = dsuid->id[15];
    } else {
        return DSUID_RC_INV_DSUID;
    }

    return DSUID_RC_OK;
}

int dsuid_get_next_dsuid(const dsuid_t *dsuid, dsuid_t *dsuid_next)
{
    uint32_t serial;

    int rv = dsuid_get_serial_number(dsuid, &serial);
    if (rv != DSUID_RC_OK) {
        return rv;
    }

    *dsuid_next = *dsuid;
    serial ++;
    if (serial == 0) {
        return DSUID_RC_INV_SERIAL;
    }

    dsuid_next->id[12] = (serial >> 24) & 0xff;
    dsuid_next->id[13] = (serial >> 16) & 0xff;
    dsuid_next->id[14] = (serial >> 8) & 0xff;
    dsuid_next->id[15] = (serial >> 0) & 0xff;

    return DSUID_RC_OK;
}

int dsuid_get_previous_dsuid(const dsuid_t *dsuid, dsuid_t *dsuid_prev)
{
    uint32_t serial;

    int rv = dsuid_get_serial_number(dsuid, &serial);
    if (rv != DSUID_RC_OK) {
        return rv;
    }

    *dsuid_prev = *dsuid;
    serial --;

    dsuid_prev->id[12] = (serial >> 24) & 0xff;
    dsuid_prev->id[13] = (serial >> 16) & 0xff;
    dsuid_prev->id[14] = (serial >> 8) & 0xff;
    dsuid_prev->id[15] = (serial >> 0) & 0xff;

    return DSUID_RC_OK;
}

int dsuid_get_main_dsuid(const dsuid_t *dsuid, const uint8_t module_count, dsuid_t *dsuid_main)
{
    uint32_t serial;

    int rv = dsuid_get_serial_number(dsuid, &serial);
    if (rv != DSUID_RC_OK) {
        return rv;
    }
    if (serial < module_count || module_count == 0) {
        return DSUID_RC_INV_SERIAL;
    }

    uint32_t offset = serial % module_count;
    uint32_t serial_main = serial - offset;
    if (serial_main % module_count != 0) {
        return DSUID_RC_INV_SERIAL;
    }

    *dsuid_main = *dsuid;
    dsuid_main->id[12] = (serial_main >> 24) & 0xff;
    dsuid_main->id[13] = (serial_main >> 16) & 0xff;
    dsuid_main->id[14] = (serial_main >> 8) & 0xff;
    dsuid_main->id[15] = (serial_main >> 0) & 0xff;

    return DSUID_RC_OK;
}

int dsuid_get_serial_number(const dsuid_t *dsuid, uint32_t *serialNumber)
{
    if ((dsuid->id[6] == 0) &&
        (dsuid->id[7] == 0) &&
        (dsuid->id[8] == 0) &&
        (dsuid->id[9] == 0))
    {
        *serialNumber = (dsuid->id[12] << 24) |
                        (dsuid->id[13] << 16) |
                        (dsuid->id[14] << 8) |
                        dsuid->id[15];
    } else {
        return DSUID_RC_INV_DSUID;
    }

    return DSUID_RC_OK;
}

int dsuid_to_string(const dsuid_t *dsuid, char *buf)
{
    char *p = buf;
    unsigned n;

    for (n = 0; n < sizeof(dsuid_t); n++) {
        p += sprintf(p, "%02x", dsuid->id[n]);
    }
    if ((p - buf) != (DSUID_STR_LEN - 1)) {
        return DSUID_RC_GEN_FAIL;
    }
    return DSUID_RC_OK;
}

__thread DSUID_STR(tls_dsuid_str);
const char *strdsuid(const dsuid_t *dsuid)
{
    dsuid_to_string(dsuid, tls_dsuid_str);
    return tls_dsuid_str;
}

static char format[] = "%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx";
int dsuid_from_string(const char *buf, dsuid_t *dsuid)
{
    int ret;
    ret = sscanf(buf, format,
            &dsuid->id[0],
            &dsuid->id[1],
            &dsuid->id[2],
            &dsuid->id[3],
            &dsuid->id[4],
            &dsuid->id[5],
            &dsuid->id[6],
            &dsuid->id[7],
            &dsuid->id[8],
            &dsuid->id[9],
            &dsuid->id[10],
            &dsuid->id[11],
            &dsuid->id[12],
            &dsuid->id[13],
            &dsuid->id[14],
            &dsuid->id[15],
            &dsuid->id[16]);

    if (ret != sizeof(dsuid->id)) {
        return DSUID_RC_GEN_FAIL;
    }
    return DSUID_RC_OK;
}

bool dsuid_is_null(const dsuid_t *dsuid)
{
    /* dsuid might be unaligned */
    return (memcmp((char *)dsuid, &DSUID_NULL, sizeof(dsuid_t)) == 0);
}

bool dsuid_equal(const dsuid_t *dsuid1, const dsuid_t *dsuid2)
{
    /* dsuid might be unaligned */
    return (memcmp((char *)dsuid1, (char *)dsuid2, sizeof(dsuid_t)) == 0);
}

bool dsuid_is_broadcast(const dsuid_t *dsuid)
{
    /* dsuid might be unaligned */
    return (memcmp((char *)dsuid, &DSUID_BROADCAST, sizeof(dsuid_t)) == 0);
}
